import random
import string

from typing import TYPE_CHECKING

import numpy as np
from vectordb.db.executors.typed_executor import TypedExecutor
from jina.serve.executors.decorators import requests, write

if TYPE_CHECKING:
    from docarray.index import HnswDocumentIndex
    from docarray import DocList


class HNSWLibIndexer(TypedExecutor):

    def __init__(self,
                 space='l2',
                 max_elements=1024,
                 ef_construction=200,
                 ef=10,
                 M=16,
                 allow_replace_deleted=False,
                 num_threads=1,
                 *args, **kwargs):
        from docarray.index import HnswDocumentIndex
        super().__init__(*args, **kwargs)
        workspace = self.workspace.replace('[', '_').replace(']', '_')
        self.work_dir = f'{workspace}' if self.handle_persistence else f'{workspace}/{"".join(random.choice(string.ascii_lowercase) for _ in range(5))}'
        db_conf = HnswDocumentIndex.DBConfig()
        db_conf.default_column_config.get(np.ndarray).update({'space': space,
                                                              'ef_construction': ef_construction,
                                                              'ef': ef,
                                                              'max_elements': max_elements,
                                                              'M': M,
                                                              'allow_replace_deleted': allow_replace_deleted,
                                                              'num_threads': num_threads})
        db_conf.work_dir = self.work_dir
        self._indexer = HnswDocumentIndex[self._input_schema](db_config=db_conf)

    def _index(self, docs, *args, **kwargs):
        self._indexer.index(docs)
        return docs

    def index(self, docs, *args, **kwargs):
        self.logger.debug(f'Index {len(docs)}')
        return self._index(docs)

    @write
    @requests(on='/index')
    async def async_index(self, docs, *args, **kwargs):
        # Index does not work on a separate thread, this is why we need to call async
        return self.index(docs)

    def search(self, docs, parameters, *args, **kwargs):
        from docarray import DocList
        self.logger.debug(f'Search')
        res = DocList[self._output_schema]()
        search_field = 'embedding'
        if parameters is not None and 'search_field' in parameters:
            search_field = parameters.pop('search_field')

        params = parameters or {}
        if '__results__' in params:
            params.pop('__results__')

        ret = self._indexer.find_batched(docs, search_field=search_field, **params)
        matched_documents = ret.documents
        matched_scores = ret.scores
        assert len(docs) == len(matched_documents) == len(matched_scores)

        for query, matches, scores in zip(docs, matched_documents, matched_scores):
            output_doc = self._output_schema(**query.dict())
            output_doc.matches = matches
            output_doc.scores = scores.tolist()
            res.append(output_doc)
        return res

    @requests(on='/search')
    async def async_search(self, docs, *args, **kwargs):
        return self.search(docs, *args, **kwargs)

    def _delete(self, docs, *args, **kwargs):
        del self._indexer[[doc.id for doc in docs]]
        return docs

    def delete(self, docs, *args, **kwargs):
        self.logger.debug(f'Delete')
        return self._delete(docs, *args, **kwargs)

    @write
    @requests(on='/delete')
    async def async_delete(self, docs, *args, **kwargs):
        return self.delete(docs, *args, **kwargs)

    def update(self, docs, *args, **kwargs):
        self.logger.debug(f'Update')
        self._delete(docs)
        return self._index(docs, *args, **kwargs)

    @write
    @requests(on='/update')
    async def async_update(self, docs, *args, **kwargs):
        # TODO: Check doc.id is present (so that it updates with shards)
        return self.update(docs, *args, **kwargs)

    def num_docs(self, **kwargs):
        return {'num_docs': self._index.num_docs()}

    def snapshot(self, snapshot_dir):
        # TODO: Maybe copy the work_dir to workspace if `handle` is False
        raise NotImplementedError('Act as not implemented')

    def restore(self, snapshot_dir):
        from docarray.index import HnswDocumentIndex
        self._indexer = HnswDocumentIndex[self._input_schema](work_dir=snapshot_dir, index_name='index')

    def close(self):
        pass



INPUTS = 'inputs'
DOCS = 'docs'
RETURN_TYPE = 'return_type'


def sort_matches_by_scores(func):
    """Method to ensure that return docs have matches sorted by score"""

    def wrapper(*args, **kwargs):
        from docarray import DocList
        res = func(*args, **kwargs)
        obj = args[0]
        if isinstance(res, DocList):
            for doc in res:
                matches = doc.matches
                scores = doc.scores
                # Combine the lists into pairs
                pairs = zip(scores, matches)

                # Sort the pairs based on the values of list1
                sorted_pairs = sorted(pairs, key=lambda x: x[0], reverse=obj.reverse_score_order)

                # Separate the sorted pairs into separate lists
                sorted_scores, sorted_matches = zip(*sorted_pairs)
                doc.matches = sorted_matches
                doc.scores = sorted_scores
        else:
            matches = res.matches
            scores = res.scores
            # Combine the lists into pairs
            pairs = zip(scores, matches)

            # Sort the pairs based on the values of list1
            sorted_pairs = sorted(pairs, key=lambda x: x[0], reverse=obj.reverse_score_order)

            # Separate the sorted pairs into separate lists
            sorted_scores, sorted_matches = zip(*sorted_pairs)
            res.matches = sorted_matches
            res.scores = sorted_scores
        return res

    return wrapper



from typing import TYPE_CHECKING

from vectordb.db.executors.typed_executor import TypedExecutor
from jina.serve.executors.decorators import requests, write

if TYPE_CHECKING:
    from docarray.index import InMemoryExactNNIndex
    from docarray import DocList


class InMemoryExactNNIndexer(TypedExecutor):

    def __init__(self, *args, **kwargs):
        from docarray.index import InMemoryExactNNIndex
        super().__init__(*args, **kwargs)
        self._index_file_path = f'{self.workspace}/index.bin' if self.handle_persistence else None
        self._indexer = InMemoryExactNNIndex[self._input_schema](index_file_path=self._index_file_path)

    def _index(self, docs, *args, **kwargs):
        self._indexer.index(docs)
        return docs

    @write
    @requests(on='/index')
    def index(self, docs, *args, **kwargs):
        self.logger.debug(f'Index {len(docs)}')
        return self._index(docs, *args, **kwargs)

    def _search(self, docs, parameters, *args, **kwargs):
        from docarray import DocList
        res = DocList[self._output_schema]()
        search_field = 'embedding'
        if parameters is not None and 'search_field' in parameters:
            search_field = parameters.pop('search_field')

        params = parameters or {}
        if '__results__' in params:
            params.pop('__results__')
        ret = self._indexer.find_batched(docs, search_field=search_field, **params)
        matched_documents = ret.documents
        matched_scores = ret.scores
        assert len(docs) == len(matched_documents) == len(matched_scores)

        for query, matches, scores in zip(docs, matched_documents, matched_scores):
            output_doc = self._output_schema(**query.dict())
            output_doc.matches = matches
            output_doc.scores = scores.tolist()
            res.append(output_doc)
        return res

    @requests(on='/search')
    def search(self, docs, *args, **kwargs):
        self.logger.debug(f'Search {len(docs)}')
        return self._search(docs, *args, **kwargs)

    def _delete(self, docs, *args, **kwargs):
        del self._indexer[[doc.id for doc in docs]]
        return docs

    @write
    @requests(on='/delete')
    def delete(self, docs, *args, **kwargs):
        self.logger.debug(f'Delete')
        return self._delete(docs, *args, **kwargs)

    @write
    @requests(on='/update')
    def update(self, docs, *args, **kwargs):
        # If not present, delete will raise IndexError and nothing will happen
        self._delete(docs)
        return self._index(docs)

    def num_docs(self, *args, **kwargs):
        return {'num_docs': self._index.num_docs()}

    def snapshot(self, snapshot_dir):
        snapshot_file = f'{snapshot_dir}/index.bin'
        self._indexer.persist(snapshot_file)

    def restore(self, snapshot_dir):
        from docarray.index import InMemoryExactNNIndex
        snapshot_file = f'{snapshot_dir}/index.bin'
        self._indexer = InMemoryExactNNIndex[self._input_schema](index_file_path=snapshot_file)

    def close(self):
        if self._index_file_path is not None:
            self._indexer.persist()






import uuid
import os
from tempfile import mktemp
import shutil
import sys
import secrets
from shutil import copytree

from pathlib import Path

__resources_path__ = os.path.join(
    Path(os.path.dirname(sys.modules['vectordb'].__file__)).parent.absolute(), 'resources'
)


class EnvironmentVarCtxtManager:
    """a class to wrap env vars"""

    def __init__(self, envs):
        """
        :param envs: a dictionary of environment variables
        """
        self._env_keys_added = envs
        self._env_keys_old = {}

    def __enter__(self):
        for key, val in self._env_keys_added.items():
            # Store the old value, if it exists
            if key in os.environ:
                self._env_keys_old[key] = os.environ[key]
            # Update the environment variable with the new value
            os.environ[key] = str(val)

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Restore the old values of updated environment variables
        for key, val in self._env_keys_old.items():
            os.environ[key] = str(val)
        # Remove any newly added environment variables
        for key in self._env_keys_added.keys():
            os.unsetenv(key)


def get_uri(id: str, tag: str):
    import requests
    from hubble import Auth

    r = requests.get(
        f"https://apihubble.jina.ai/v2/executor/getMeta?id={id}&tag={tag}",
        headers={"Authorization": f"token {Auth.get_auth_token()}"},
    )
    _json = r.json()
    if _json is None:
        print(f'Could not find image with id {id} and tag {tag}')
        return
    _image_name = _json['data']['name']
    _user_name = _json['meta']['owner']['name']
    return f'jinaai+docker://{_user_name}/{_image_name}:{tag}'


def get_random_tag():
    return 't-' + uuid.uuid4().hex[:5]


def get_random_name():
    return 'n-' + uuid.uuid4().hex[:5]


def _push_to_hubble(
        tmpdir: str, name: str, tag: str, verbose: bool, public: bool
) -> str:
    from hubble.executor.hubio import HubIO
    from hubble.executor.parsers import set_hub_push_parser

    secret = secrets.token_hex(8)
    args_list = [
        tmpdir,
        '--tag',
        tag,
        '--no-usage',
        '--no-cache',
    ]
    if verbose:
        args_list.remove('--no-usage')
        args_list.append('--verbose')
    if not public:
        args_list.append('--secret')
        args_list.append(secret)
        args_list.append('--private')

    args = set_hub_push_parser().parse_args(args_list)

    push_envs = (
        {'JINA_HUBBLE_HIDE_EXECUTOR_PUSH_SUCCESS_MSG': 'true'} if not verbose else {}
    )
    # return 'a' + ':' + tag

    with EnvironmentVarCtxtManager(push_envs):
        executor_id = HubIO(args).push().get('id')
        return executor_id + ':' + tag


def push_vectordb_to_hubble(
        vectordb_name,
        definition_file_path,
) -> str:
    tmpdir = mktemp()
    image_name = get_random_name()
    tag = get_random_name()
    copytree(os.path.join(__resources_path__, 'jcloud_exec_template'), tmpdir)
    shutil.copy(definition_file_path, os.path.join(tmpdir, 'vectordb_app.py'))
    # replace `vectordb_name` in `vectordb_app`
    with open(os.path.join(tmpdir, 'executor.py'), encoding='utf-8') as f:
        content = f.read()

    content = content.replace('from vectordb_app import db', f'from vectordb_app import {vectordb_name}')
    content = content.replace('class VectorDBExecutor(db._executor_cls):',
                              f'class VectorDBExecutor({vectordb_name}._executor_cls):')

    with open(os.path.join(tmpdir, 'executor.py'), mode='w', encoding='utf-8') as f:
        f.write(content)

    with open(os.path.join(tmpdir, 'config.yml'), encoding='utf-8') as f:
        content = f.read()

    content = content.replace('vectordb_executor', f'vectordb_executor-{image_name}')

    with open(os.path.join(tmpdir, 'config.yml'), mode='w', encoding='utf-8') as f:
        f.write(content)

    executor_id = _push_to_hubble(tmpdir, image_name, tag, True, False)
    id, tag = executor_id.split(':')
    return get_uri(id, tag)



from vectordb.client.client import Client
from vectordb.utils.unify_input_output import unify_input_output
from vectordb.utils.pass_parameters import pass_kwargs_as_params


class Service:

    def __init__(self, ctxt_manager, schema, address, reverse_order=False):
        self.ctxt_manager = ctxt_manager
        self._reverse_order = reverse_order
        self._client = Client[schema](address, reverse_order=reverse_order)

    def __enter__(self):
        self.ctxt_manager.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.ctxt_manager.__exit__(exc_type, exc_val, exc_tb)

    @property
    def reverse_score_order(self):
        return self._reverse_order

    def block(self):
        return self.ctxt_manager.block()

    def client(self):
        return self._client

    @pass_kwargs_as_params
    @unify_input_output
    def index(self, *args, **kwargs):
        return self._client.index(*args, **kwargs)

    @pass_kwargs_as_params
    @unify_input_output
    def search(self, *args, **kwargs):
        return self._client.search(*args, **kwargs)

    @pass_kwargs_as_params
    @unify_input_output
    def delete(self, *args, **kwargs):
        return self._client.delete(*args, **kwargs)

    @pass_kwargs_as_params
    @unify_input_output
    def update(self, *args, **kwargs):
        return self._client.update(*args, **kwargs)

    @pass_kwargs_as_params
    @unify_input_output
    def post(self, *args, **kwargs):
        return self._client.post(*args, **kwargs)



from typing import TypeVar, Generic, Type, Optional, Union, List, Dict, TYPE_CHECKING
from vectordb.db.executors.typed_executor import TypedExecutor
from vectordb.db.service import Service
from vectordb.utils.unify_input_output import unify_input_output
from vectordb.utils.pass_parameters import pass_kwargs_as_params
from vectordb.utils.sort_matches_by_score import sort_matches_by_scores
from vectordb.utils.push_to_hubble import push_vectordb_to_hubble

if TYPE_CHECKING:
    from jina import Deployment, Flow
    from docarray import BaseDoc, DocList

TSchema = TypeVar('TSchema', bound='BaseDoc')

REQUESTS_MAP = {'/index': 'index', '/update': 'update', '/delete': 'delete', '/search': 'search'}


class VectorDB(Generic[TSchema]):
    # the BaseDoc that defines the schema of the store
    # for subclasses this is filled automatically
    _input_schema: Optional[Type['BaseDoc']] = None
    _executor_type: Optional[Type[TypedExecutor]] = None
    _executor_cls: Type[TypedExecutor]

    ##################################################
    # Behind-the-scenes magic                        #
    # Subclasses should not need to implement these  #
    ##################################################
    def __class_getitem__(cls, item: Type[TSchema]):
        from docarray import BaseDoc
        if not isinstance(item, type):
            # do nothing
            # enables use in static contexts with type vars, e.g. as type annotation
            return Generic.__class_getitem__.__func__(cls, item)
        if not issubclass(item, BaseDoc):
            raise ValueError(
                f'{cls.__name__}[item] `item` should be a Document not a {item} '
            )

        class VectorDBTyped(cls):  # type: ignore
            _input_schema: Type[TSchema] = item
            _executor_cls: Type[TypedExecutor] = cls._executor_type[item]

        VectorDBTyped.__name__ = f'{cls.__name__}[{item.__name__}]'
        VectorDBTyped.__qualname__ = f'{cls.__qualname__}[{item.__name__}]'

        return VectorDBTyped

    def __init__(self, *args, **kwargs):
        self._workspace = None
        if 'work_dir' in kwargs:
            self._workspace = kwargs['work_dir']
        if 'workspace' in kwargs:
            self._workspace = kwargs['workspace']
        self._uses_with = kwargs
        kwargs['requests'] = REQUESTS_MAP
        kwargs['runtime_args'] = {'workspace': self._workspace}
        self._executor = self._executor_cls(*args, **kwargs)

    @classmethod
    def _get_jina_object(cls,
                         *,
                         to_deploy: bool = False,
                         port: Optional[Union[str, List[str]]] = 8081,
                         protocol: Optional[Union[str, List[str]]] = None,
                         workspace: Optional[str] = None,
                         shards: Optional[int] = None,
                         replicas: Optional[int] = None,
                         peer_ports: Optional[Union[Dict[str, List], List]] = None,
                         uses_with: Optional[Dict] = None,
                         definition_file: Optional[str] = None,
                         obj_name: Optional[str] = None,
                         **kwargs):
        from jina import Deployment, Flow
        is_instance = False
        uses_with = uses_with or {}
        if isinstance(cls, VectorDB):
            is_instance = True
            uses_with = uses_with.update(**cls._uses_with)

        if is_instance:
            workspace = workspace or cls._workspace
        replicas = replicas or 1
        shards = shards or 1
        protocol = protocol or 'grpc'
        protocol_list = [p.lower() for p in protocol] if isinstance(protocol, list) else [protocol.lower()]
        stateful = replicas is not None and replicas > 1
        if not to_deploy:
            executor_cls_name = ''.join(cls._executor_cls.__name__.split('[')[0:2])
            ServedExecutor = type(f'{executor_cls_name.replace("[", "").replace("]", "")}', (cls._executor_cls,),
                                  {})
            uses = ServedExecutor
        polling = {'/index': 'ANY', '/search': 'ALL', '/update': 'ALL', '/delete': 'ALL'}
        if to_deploy and replicas > 1:
            import warnings
            warnings.warn(
                'Deployment with replicas > 1 is not currently available. The deployment will have 1 replica per each '
                'shard')
            replicas = 1

        if 1 < replicas < 3:
            raise Exception(f'In order for consensus to properly work, at least 3 replicas need to be provided.')

        if peer_ports is None and stateful is True:
            peer_ports = {}
            if shards is not None:
                for shard in range(shards):
                    peer_ports[str(shard)] = []
                    for replica in range(replicas):
                        peer_ports[str(shard)].append(port + shard * replicas + replica + 1)
            else:
                peer_ports['0'] = [port + (replica + 1) for replica in range(replicas)]

        if stateful is True:
            if shards is not None:
                assert len(peer_ports.keys()) == shards, 'Need to provide `peer_ports` information for each shard'
            for shard in peer_ports.keys():
                assert len(peer_ports[str(
                    shard)]) == replicas, f'For shard {shard}, need to provide ports for each replica ({replicas} replicas), got {len(peer_ports[str(shard)])} instead'

        if 'stateful' in kwargs:
            kwargs.pop('stateful')

        use_deployment = True
        if to_deploy:
            # here we would need to push the EXECUTOR TO HUBBLE AND CHANGE THE USES
            assert definition_file is not None, 'Trying to create a Jina Object for Deployment without the file where the vectordb object/class is defined'
            assert obj_name is not None, 'Trying to create a Jina Object for Deployment without the name of the vectordb object/class to deploy'
            uses = f'{push_vectordb_to_hubble(vectordb_name=obj_name, definition_file_path=definition_file)}'
            use_deployment = False
            port = 8080

        if 'websocket' in protocol_list:  # websocket not supported for Deployment
            use_deployment = False

        if (shards > 1 or stateful) and 'http' in protocol_list:  # http not supported for shards > 1 or stateful
            use_deployment = False

        if use_deployment:
            jina_object = Deployment(name='indexer',
                                     uses=uses,
                                     uses_with=uses_with,
                                     port=port,
                                     protocol=protocol,
                                     shards=shards,
                                     replicas=replicas,
                                     stateful=stateful,
                                     peer_ports=peer_ports,
                                     workspace=workspace,
                                     polling=polling, **kwargs)
        else:
            jina_object = Flow(port=port, protocol=protocol, env=['JINA_LOG_LEVEL=DEBUG'], **kwargs).add(name='indexer',
                                                                           uses=uses,
                                                                           uses_with=uses_with,
                                                                           shards=shards,
                                                                           replicas=replicas,
                                                                           stateful=stateful,
                                                                           peer_ports=peer_ports,
                                                                           polling=polling,
                                                                           workspace=workspace)

        return jina_object

    @classmethod
    def serve(cls,
              *,
              port: Optional[Union[str, List[str]]] = 8081,
              protocol: Optional[Union[str, List[str]]] = None,
              **kwargs):
        protocol = protocol or 'grpc'
        protocol_list = [p.lower() for p in protocol] if isinstance(protocol, list) else [protocol.lower()]
        ctxt_manager = cls._get_jina_object(to_deploy=False, port=port, protocol=protocol, **kwargs)
        port = port[0] if isinstance(port, list) else port
        return Service(ctxt_manager, address=f'{protocol_list[0]}://0.0.0.0:{port}', schema=cls._input_schema,
                       reverse_order=cls.reverse_score_order)

    @classmethod
    def deploy(cls,
               **kwargs):
        from tempfile import mkdtemp
        import os
        import yaml
        from yaml.loader import SafeLoader
        jina_obj = cls._get_jina_object(to_deploy=True, **kwargs)

        tmpdir = mkdtemp()
        jina_obj.save_config(os.path.join(tmpdir, 'flow.yml'))
        with open(os.path.join(tmpdir, 'flow.yml'), 'r') as f:
            flow_dict = yaml.load(f, SafeLoader)

        executor_jcloud_config = {'resources': {'instance': 'C5', 'autoscale': {'min': 0, 'max': 1},
                                                'storage': {'kind': 'ebs', 'size': '10G', 'retain': True}}}
        for executor in flow_dict['executors']:
            executor['jcloud'] = executor_jcloud_config

        global_jcloud_config = {
            'docarray': '0.34.0',
            'labels': {
                'app': 'vectordb',
            },
            'monitor': {
                'traces': {
                    'enable': True,
                },
                'metrics': {
                    'enable': True,
                    'host': 'http://opentelemetry-collector.monitor.svc.cluster.local',
                    'port': 4317,
                },
            },
        }
        flow_dict['jcloud'] = global_jcloud_config
        import tempfile
        from jcloud.flow import CloudFlow

        with tempfile.TemporaryDirectory() as tmpdir:
            flow_path = os.path.join(tmpdir, 'flow.yml')
            print(f' flow_path {flow_path}')
            with open(flow_path, 'w') as f:
                yaml.safe_dump(flow_dict, f, sort_keys=False)

            cloud_flow = CloudFlow(path=flow_path)

            async def _deploy():
                await cloud_flow.__aenter__()

            import asyncio
            ret = asyncio.run(_deploy())
        return ret

    @pass_kwargs_as_params
    @unify_input_output
    def index(self, docs: 'DocList[TSchema]', parameters: Optional[Dict] = None, **kwargs):
        return self._executor.index(docs, parameters)

    @pass_kwargs_as_params
    @unify_input_output
    def update(self, docs: 'DocList[TSchema]', parameters: Optional[Dict] = None, **kwargs):
        return self._executor.update(docs, parameters)

    @pass_kwargs_as_params
    @unify_input_output
    def delete(self, docs: 'DocList[TSchema]', parameters: Optional[Dict] = None, **kwargs):
        return self._executor.delete(docs, parameters)

    @sort_matches_by_scores
    @pass_kwargs_as_params
    @unify_input_output
    def search(self, docs: 'DocList[TSchema]', parameters: Optional[Dict] = None, **kwargs):
        return self._executor.search(docs, parameters)

    def persist(self):
        return self._executor.close()



import inspect


INPUTS = 'inputs'
DOCS = 'docs'


def unify_input_output(func):
    """Method to ensure the input docs are passed internally as a list, and that the return docs are returned as
    single if needed. It also makes sure that it unifies `docs` and `inputs` usage"""

    func_args = inspect.getfullargspec(func).args

    expected_key = DOCS if DOCS in func_args else INPUTS

    def wrapper(*args, **kwargs):
        from docarray import DocList, BaseDoc
        return_single = False
        present_key = None
        if DOCS in kwargs:
            present_key = DOCS
            docs = kwargs[DOCS]
        elif INPUTS in kwargs:
            present_key = INPUTS
            docs = kwargs[INPUTS]
        else:
            docs = args[1]

        new_args = args
        if isinstance(docs, BaseDoc):
            return_single = True
            docs = DocList.__class_getitem__(docs.__class__)([docs])
            if present_key is None:
                new_args = (args[0], docs)
            else:
                kwargs[expected_key] = docs
        elif present_key is not None and expected_key != present_key:
            kwargs[expected_key] = docs

        ret = func(*new_args, **kwargs)
        if return_single:
            return ret[0]
        else:
            return ret

    return wrapper



def _ignore_warnings():
    import logging
    import warnings

    logging.captureWarnings(True)
    warnings.filterwarnings(
        "ignore",
        category=DeprecationWarning,
        message="Deprecated call to `pkg_resources.declare_namespace('google')`.",
    )


_ignore_warnings()

__version__ = '0.0.15.dev2'

from vectordb.client import Client
from vectordb.db.hnsw_vectordb import HNSWVectorDB
from vectordb.db.inmemory_exact_vectordb import InMemoryExactNNVectorDB



import click

from vectordb.db.base import VectorDB
from vectordb import __version__


@click.group()
@click.version_option(__version__, '-v', '--version', prog_name='vectordb')
@click.help_option('-h', '--help')
def vectordb():
    pass


@vectordb.command(help='Deploy a vectorDB db to Jina AI Cloud')
@click.option(
    '--db',
    '--app',
    type=str,
    required=True,
    help='VectorDB to deploy, in the format "<module>:<attribute>"',
)
@click.option(
    '--protocol',
    '--protocols',
    type=str,
    default='grpc',
    help='Protocol to use to communicate with the VectorDB. It can be a single one or a list of multiple protocols to use. Options are grpc, http and websocket',
    required=False,
    show_default=True,
)
@click.option(
    '--shards',
    '-s',
    type=int,
    default=1,
    help='Number of shards to use for the deployed VectorDB',
    required=False,
    show_default=True,
)
def deploy(db, protocol, shards):
    definition_file, _, obj_name = db.partition(":")
    if not definition_file.endswith('.py'):
        definition_file = f'{definition_file}.py'
    protocol = protocol.split(',')
    VectorDB.deploy(protocol=protocol,
                    shards=shards,
                    definition_file=definition_file,
                    obj_name=obj_name)


@vectordb.command(help='Locally serve a vectorDB db')
@click.option(
    '--db',
    '--app',
    type=str,
    required=True,
    help='VectorDB to serve, in the format "<module>:<attribute>"',
)
@click.option(
    '--port',
    '-p',
    type=str,
    default='8081',
    help='Port to use to access the VectorDB. It can be a single one or a list corresponding to the protocols',
    required=False,
    show_default=True,
)
@click.option(
    '--protocol',
    '--protocols',
    type=str,
    default='grpc',
    help='Protocol to use to communicate with the VectorDB. It can be a single one or a list of multiple protocols to use. Options are grpc, http and websocket',
    required=False,
    show_default=True,
)
@click.option(
    '--replicas',
    '-r',
    type=int,
    default=1,
    help='Number of replicas to use for the serving VectorDB',
    required=False,
    show_default=True,
)
@click.option(
    '--shards',
    '-s',
    type=int,
    default=1,
    help='Number of shards to use for the served VectorDB',
    required=False,
    show_default=True,
)
@click.option(
    '--workspace',
    '-w',
    type=str,
    default='.',
    help='Workspace for the VectorDB to persist its data',
    required=False,
    show_default=True,
)
def serve(db, port, protocol, replicas, shards, workspace):
    import importlib
    definition_file, _, obj_name = db.partition(":")
    port = port.split(',')
    if len(port) == 1:
        port = int(port[0])
    else:
        port = [int(p) for p in port]
    protocol = protocol.split(',')
    if definition_file.endswith('.py'):
        definition_file = definition_file[:-3]
    module = importlib.import_module(definition_file)
    db = getattr(module, obj_name)
    service = db.serve(protocol=protocol,
                       port=port,
                       shards=shards,
                       replicas=replicas,
                       workspace=workspace)
    with service:
        service.block()


if __name__ == '__main__':
    vectordb()



from vectordb.db.base import VectorDB
from vectordb.db.executors.hnsw_indexer import HNSWLibIndexer


class HNSWVectorDB(VectorDB):
    _executor_type = HNSWLibIndexer
    reverse_score_order = False



from typing import Type, List, TYPE_CHECKING

if TYPE_CHECKING:
    from docarray import BaseDoc


def create_output_doc_type(input_doc_type: Type['BaseDoc']):
    """
    Given a Document type, dynamically generate a new `BaseDoc` that will include nested matches of input doc type as matches and a list of floats as scores of those matches
    :param input_doc_type: The document type
    :return: The created output type
    """
    from docarray import BaseDoc, DocList
    from pydantic import create_model
    return create_model(
        input_doc_type.__name__ + 'WithMatchesAndScores',
        __base__=input_doc_type,
        __validators__=input_doc_type.__validators__,
        matches=(DocList[input_doc_type], []),
        scores=(List[float], [])
    )



from vectordb.db.base import VectorDB
from vectordb.db.executors.inmemory_exact_indexer import InMemoryExactNNIndexer


class InMemoryExactNNVectorDB(VectorDB):
    _executor_type = InMemoryExactNNIndexer
    reverse_score_order = True



from jina import Executor
from jina.serve.executors import _FunctionWithSchema
from jina.serve.executors import __dry_run_endpoint__

from typing import TypeVar, Generic, Type, Optional, TYPE_CHECKING
from vectordb.utils.create_doc_type import create_output_doc_type


if TYPE_CHECKING:
    from docarray import BaseDoc, DocList

TSchema = TypeVar('TSchema', bound='BaseDoc')

InputSchema = TypeVar('InputSchema', bound='BaseDoc')
OutputSchema = TypeVar('OutputSchema', bound='BaseDoc')

methods = ['/index', '/update', '/delete', '/search']


class TypedExecutor(Executor, Generic[InputSchema, OutputSchema]):
    # the BaseDoc that defines the schema of the store
    # for subclasses this is filled automatically
    _input_schema: Optional[Type['BaseDoc']] = None
    _output_schema: Optional[Type['BaseDoc']] = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        from docarray import DocList
        self._num_replicas = getattr(self.runtime_args, 'replicas', 1)
        for k, v in self._requests.items():
            if k != __dry_run_endpoint__:
                if k != '/search':
                    self._requests[k] = _FunctionWithSchema(self._requests[k].fn, DocList[self._input_schema],
                                                            DocList[self._input_schema])
                else:
                    self._requests[k] = _FunctionWithSchema(self._requests[k].fn, DocList[self._input_schema],
                                                            DocList[self._output_schema])

    @property
    def handle_persistence(self):
        return self._num_replicas == 1

    ##################################################
    # Behind-the-scenes magic                        #
    # Subclasses should not need to implement these  #
    ##################################################
    def __class_getitem__(cls, item: Type[InputSchema]):

        from docarray import BaseDoc
        if not isinstance(item, type):
            # do nothing
            # enables use in static contexts with type vars, e.g. as type annotation
            return Generic.__class_getitem__.__func__(cls, item)
        if not issubclass(item, BaseDoc):
            raise ValueError(
                f'{cls.__name__}[item] `item` should be a Document not a {item} '
            )

        input_schema = item
        output_schema = create_output_doc_type(input_schema)

        class ExecutorTyped(cls):  # type: ignore
            _input_schema: Type[InputSchema] = input_schema
            _output_schema: Type[OutputSchema] = output_schema

        ExecutorTyped.__name__ = f'{cls.__name__}[{input_schema.__name__}][{output_schema.__name__}]'
        ExecutorTyped.__qualname__ = f'{cls.__qualname__}[{input_schema.__name__}][{output_schema.__name__}]'

        return ExecutorTyped



from typing import TypeVar, Generic, Type, Optional, TYPE_CHECKING
from vectordb.utils.unify_input_output import unify_input_output
from vectordb.utils.pass_parameters import pass_kwargs_as_params
from vectordb.utils.create_doc_type import create_output_doc_type
from vectordb.utils.sort_matches_by_score import sort_matches_by_scores


if TYPE_CHECKING:
    from docarray import BaseDoc, DocList

TSchema = TypeVar('TSchema', bound='BaseDoc')

REQUESTS_MAP = {'/index': 'index', '/update': 'update', '/delete': 'delete', '/search': 'search'}


class Client(Generic[TSchema]):

    _input_schema: Optional[Type['BaseDoc']] = None
    _output_schema: Optional[Type['BaseDoc']] = None

    ##################################################
    # Behind-the-scenes magic                        #
    # Subclasses should not need to implement these  #
    ##################################################
    def __class_getitem__(cls, item: Type[TSchema]):
        from docarray import BaseDoc
        if not isinstance(item, type):
            # do nothing
            # enables use in static contexts with type vars, e.g. as type annotation
            return Generic.__class_getitem__.__func__(cls, item)
        if not issubclass(item, BaseDoc):
            raise ValueError(
                f'{cls.__name__}[item] `item` should be a Document not a {item} '
            )

        out_item = create_output_doc_type(item)

        class ClientTyped(cls):  # type: ignore
            _input_schema: Type[TSchema] = item
            _output_schema: Type[TSchema] = out_item

        ClientTyped.__name__ = f'{cls.__name__}[{item.__name__}]'
        ClientTyped.__qualname__ = f'{cls.__qualname__}[{item.__name__}]'

        return ClientTyped

    def __init__(self, address, reverse_order=False):
        from jina import Client as jClient
        self.reverse_score_order = reverse_order
        self._client = jClient(host=address)

    @unify_input_output
    @pass_kwargs_as_params
    def index(self, *args, **kwargs):
        return self._client.index(*args, **kwargs)

    @sort_matches_by_scores
    @unify_input_output
    @pass_kwargs_as_params
    def search(self, *args, **kwargs):
        from docarray import DocList
        return self._client.search(return_type=DocList[self._output_schema], *args, **kwargs)

    @unify_input_output
    @pass_kwargs_as_params
    def delete(self, *args, **kwargs):
        return self._client.delete(*args, **kwargs)

    @unify_input_output
    @pass_kwargs_as_params
    def update(self, *args, **kwargs):
        return self._client.update(*args, **kwargs)

    @unify_input_output
    @pass_kwargs_as_params
    def post(self, *args, **kwargs):
        return self._client.index(*args, **kwargs)



INPUTS = 'inputs'
DOCS = 'docs'
RETURN_TYPE = 'return_type'


def pass_kwargs_as_params(func):
    """Method to ensure that kwargs are passed as parameters to underlying Executor method"""

    def wrapper(*args, **kwargs):
        parameters = None
        if 'parameters' in kwargs:
            parameters = kwargs.pop('parameters')

        params = parameters or {}
        for k, v in kwargs.items():
            if k not in {INPUTS, DOCS, RETURN_TYPE}:
                params[k] = kwargs[k]
        if len(params.keys()) > 0:
            for k in params.keys():
                if k in kwargs:
                    kwargs.pop(k)
            kwargs['parameters'] = params

        return func(*args, **kwargs)

    return wrapper



from setuptools import setup, find_packages
from os import path

try:
    pkg_name = 'vectordb'
    libinfo_py = path.join(pkg_name, '__init__.py')
    libinfo_content = open(libinfo_py, 'r', encoding='utf-8').readlines()
    version_line = [l.strip() for l in libinfo_content if l.startswith('__version__')][
        0
    ]
    exec(version_line)  # gives __version__
except FileNotFoundError:
    __version__ = '0.0.0'

try:
    with open('README.md', encoding='utf-8') as fp:
        _long_description = fp.read()
except FileNotFoundError:
    _long_description = ''


# Read the contents of requirements.txt
with open(path.join(path.dirname(__file__), 'requirements.txt'), 'r') as f:
    requirements = f.read().splitlines()

setup(
    name='vectordb',
    version=__version__,
    description='The Python VectorDB. Build your vector database from working as a library to scaling as a database in the cloud',
    long_description=_long_description,
    long_description_content_type='text/markdown',
    author='Jina AI',
    author_email='hello@jina.ai',
    license='Apache 2.0',
    url='https://github.com/jina-ai/vectordb/',
    download_url='https://github.com/jina-ai/vectordb/tags',
    packages=find_packages(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    python_requires='>=3.7',
    entry_points={
        'console_scripts': [
            'vectordb=vectordb.__main__:vectordb',
        ],
    },
    extras_require={
        'test': [
            'pytest',
            'pytest-asyncio',
            'pytest-repeat',
            'flaky',
            'pytest-timeout'
        ],
    },
    install_requires=requirements,
)
